#Gets model predictions and saves them to a CSV file to prevent having to get predicitions every time a metric is calculated.

import pandas as pd
import numpy as np
from tqdm.auto import tqdm
import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
from transformers import BertTokenizerFast, AutoTokenizer, BertModel, get_linear_schedule_with_warmup, BertForSequenceClassification
import pytorch_lightning as pl
from torch.optim import AdamW
from torchmetrics.functional import accuracy, f1, auroc
from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping
from pytorch_lightning.loggers import TensorBoardLogger
from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report, multilabel_confusion_matrix
from sklearn import metrics
import seaborn as sns
from pylab import rcParams
import os

#os.environ["CUDA_VISIBLE_DEVICES"]="1"
#os.environ["no_deprecation_warning"]="True"

LABEL_COLUMNS = ['label']
max_token_len = 1024



class norBERTLongSelfAttention(LongformerSelfAttention):
    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_value=None,
        output_attentions=False,
    ):
        """
        :class:`LongformerSelfAttention` expects `len(hidden_states)` to be multiple of `attention_window`. Padding to
        `attention_window` happens in :meth:`LongformerModel.forward` to avoid redoing the padding on each layer.

        Modified LongformerSelfAttention layer to work with RoBERTa models. They have different set of arguments
        compared to original Longformer attention: RoBERTa is passing some additional arguments and arguments required
        by LongformerSelfAttention are missing so we have to add it manually.

        The `attention_mask` is changed in :meth:`LongformerModel.forward` from 0, 1, 2 to:

            * -10000: no attention
            * 0: local attention
            * +10000: global attention
        """
        hidden_states = hidden_states.transpose(0, 1)

        # project hidden states
        query_vectors = self.query(hidden_states)
        key_vectors = self.key(hidden_states)
        value_vectors = self.value(hidden_states)

        ## Lines below has been changed

        attention_mask = attention_mask.squeeze(dim=2).squeeze(dim=1)

        # is index masked or global attention
        is_index_masked = attention_mask < 0
        is_index_global_attn = attention_mask > 0
        is_global_attn = any(is_index_global_attn.flatten())

        ## End of modification

        seq_len, batch_size, embed_dim = hidden_states.size()
        assert (
            embed_dim == self.embed_dim
        ), f"hidden_states should have embed_dim = {self.embed_dim}, but has {embed_dim}"

        # normalize query
        query_vectors /= math.sqrt(self.head_dim)

        query_vectors = query_vectors.view(seq_len, batch_size, self.num_heads, self.head_dim).transpose(0, 1)
        key_vectors = key_vectors.view(seq_len, batch_size, self.num_heads, self.head_dim).transpose(0, 1)

        attn_scores = self._sliding_chunks_query_key_matmul(
            query_vectors, key_vectors, self.one_sided_attn_window_size
        )

        # values to pad for attention probs
        
        ## Lines below has been changed
        
        remove_from_windowed_attention_mask = (attention_mask != 0).unsqueeze(dim=-1).unsqueeze(dim=-1)

        ## End of modification

        # cast to fp32/fp16 then replace 1's with -inf
        float_mask = remove_from_windowed_attention_mask.type_as(query_vectors).masked_fill(
            remove_from_windowed_attention_mask, -10000.0
        )
        # diagonal mask with zeros everywhere and -inf inplace of padding
        diagonal_mask = self._sliding_chunks_query_key_matmul(
            float_mask.new_ones(size=float_mask.size()), float_mask, self.one_sided_attn_window_size
        )

        # pad local attention probs
        attn_scores += diagonal_mask

        assert list(attn_scores.size()) == [
            batch_size,
            seq_len,
            self.num_heads,
            self.one_sided_attn_window_size * 2 + 1,
        ], f"local_attn_probs should be of size ({batch_size}, {seq_len}, {self.num_heads}, {self.one_sided_attn_window_size * 2 + 1}), but is of size {attn_scores.size()}"

        # compute local attention probs from global attention keys and contact over window dim
        if is_global_attn:
            # compute global attn indices required through out forward fn
            (
                max_num_global_attn_indices,
                is_index_global_attn_nonzero,
                is_local_index_global_attn_nonzero,
                is_local_index_no_global_attn_nonzero,
            ) = self._get_global_attn_indices(is_index_global_attn)
            # calculate global attn probs from global key

            global_key_attn_scores = self._concat_with_global_key_attn_probs(
                query_vectors=query_vectors,
                key_vectors=key_vectors,
                max_num_global_attn_indices=max_num_global_attn_indices,
                is_index_global_attn_nonzero=is_index_global_attn_nonzero,
                is_local_index_global_attn_nonzero=is_local_index_global_attn_nonzero,
                is_local_index_no_global_attn_nonzero=is_local_index_no_global_attn_nonzero,
            )
            # concat to local_attn_probs
            # (batch_size, seq_len, num_heads, extra attention count + 2*window+1)
            attn_scores = torch.cat((global_key_attn_scores, attn_scores), dim=-1)

            # free memory
            del global_key_attn_scores

        attn_probs = F.softmax(attn_scores, dim=-1, dtype=torch.float32)  # use fp32 for numerical stability

        # softmax sometimes inserts NaN if all positions are masked, replace them with 0
        attn_probs = torch.masked_fill(attn_probs, is_index_masked[:, :, None, None], 0.0)
        attn_probs = attn_probs.type_as(attn_scores)

        # free memory
        del attn_scores

        # apply dropout
        attn_probs = F.dropout(attn_probs, p=self.dropout, training=self.training)

        value_vectors = value_vectors.view(seq_len, batch_size, self.num_heads, self.head_dim).transpose(0, 1)

        # compute local attention output with global attention value and add
        if is_global_attn:
            # compute sum of global and local attn
            attn_output = self._compute_attn_output_with_global_indices(
                value_vectors=value_vectors,
                attn_probs=attn_probs,
                max_num_global_attn_indices=max_num_global_attn_indices,
                is_index_global_attn_nonzero=is_index_global_attn_nonzero,
                is_local_index_global_attn_nonzero=is_local_index_global_attn_nonzero,
            )
        else:
            # compute local attn only
            attn_output = self._sliding_chunks_matmul_attn_probs_value(
                attn_probs, value_vectors, self.one_sided_attn_window_size
            )

        assert attn_output.size() == (batch_size, seq_len, self.num_heads, self.head_dim), "Unexpected size"
        attn_output = attn_output.transpose(0, 1).reshape(seq_len, batch_size, embed_dim).contiguous()

        # compute value for global attention and overwrite to attention output
        # TODO: remove the redundant computation
        if is_global_attn:
            global_attn_output, global_attn_probs = self._compute_global_attn_output_from_hidden(
                hidden_states=hidden_states,
                max_num_global_attn_indices=max_num_global_attn_indices,
                is_local_index_global_attn_nonzero=is_local_index_global_attn_nonzero,
                is_index_global_attn_nonzero=is_index_global_attn_nonzero,
                is_local_index_no_global_attn_nonzero=is_local_index_no_global_attn_nonzero,
                is_index_masked=is_index_masked,
            )

            # get only non zero global attn output
            nonzero_global_attn_output = global_attn_output[
                is_local_index_global_attn_nonzero[0], :, is_local_index_global_attn_nonzero[1]
            ]

            # overwrite values with global attention
            attn_output[is_index_global_attn_nonzero[::-1]] = nonzero_global_attn_output.view(
                len(is_local_index_global_attn_nonzero[0]), -1
            )
            # The attention weights for tokens with global attention are
            # just filler values, they were never used to compute the output.
            # Fill with 0 now, the correct values are in 'global_attn_probs'.
            attn_probs[is_index_global_attn_nonzero] = 0

        outputs = (attn_output.transpose(0, 1),)

        if output_attentions:
            outputs += (attn_probs,)

        return outputs + (global_attn_probs,) if (is_global_attn and output_attentions) else outputs



class norBERTLongForSequenceClassification(BertForSequenceClassification):
    def __init__(self, config):
        super().__init__(config)
        #print(nn.ModuleList([norBERTLongSelfAttention(config, layer_id=index)]))
        #self.bert.encoder.layer = nn.ModuleList([norBERTLongSelfAttention(config, layer_id=index) for index in range(config.num_hidden_layers)])
        for index in range(config.num_hidden_layers):
            norBERTLongSelfAttention(config, layer_id=index)
            

class norBERTLongModel(BertModel):
    def __init__(self, config):
        super().__init__(config)
        self.encoder.layer = nn.ModuleList([norBERTLongSelfAttention(config, layer_id=index) for index in range(config.num_hidden_layers)])






class IronicCommentModule(pl.LightningModule):

  def __init__(self, n_classes: int, n_training_steps=None, n_warmup_steps=None):
    super().__init__()

    self.bert = norBERTLongForSequenceClassification.from_pretrained('D:/noBert/LongNorBERT_v2', attention_window = 512, num_labels=1, return_dict=True)
    #self.bert = AutoModelForSequenceClassification.from_pretrained('D:/noBert/LongNorBERT', attention_window = 512, return_dict=True)
    self.classifier = nn.Linear(self.bert.config.hidden_size, n_classes)
    self.n_training_steps = n_training_steps
    self.n_warmup_steps = n_warmup_steps
    self.criterion = nn.BCELoss()



  def forward(self, input_ids, attention_mask, label=None):

    output = self.bert(input_ids, attention_mask=attention_mask)

    #Considering removing dtype
    output = torch.sigmoid(output[0])

    
    loss = 0
    if label is not None:
        loss = self.criterion(output, label)
        #loss.requres_grad = True

    return loss, output

  def training_step(self, batch, batch_idx):
    input_ids = batch["input_ids"]
    attention_mask = batch["attention_mask"]
    label = batch["label"]
    loss, outputs = self(input_ids, attention_mask, label)
    self.log("train_loss", loss, prog_bar=True, logger=True)
    return {"loss": loss, "predictions": outputs.detach(), "label": label}

  def validation_step(self, batch, batch_idx):
    input_ids = batch["input_ids"]
    attention_mask = batch["attention_mask"]
    label = batch["label"]
    loss, outputs = self(input_ids, attention_mask, label)
    self.log("val_loss", loss, prog_bar=True, logger=True)
    return loss

  def test_step(self, batch, batch_idx):
    input_ids = batch["input_ids"]
    attention_mask = batch["attention_mask"]
    label = batch["label"]
    loss, outputs = self(input_ids, attention_mask, label)
    self.log("test_loss", loss, prog_bar=True, logger=True)
    return loss

  def training_epoch_end(self, outputs):
    
    label = []
    predictions = []
    for output in outputs:
      for out_label in output["label"].detach().cpu():
        label.append(out_label)
      for out_predictions in output["predictions"].detach().cpu():
        predictions.append(out_predictions)

    label = torch.stack(label).int()
    predictions = torch.stack(predictions)

    for i, name in enumerate(LABEL_COLUMNS):
      class_roc_auc = auroc(predictions[:, i], label[:, i])
      self.logger.experiment.add_scalar(f"{name}_roc_auc/Train", class_roc_auc, self.current_epoch)


  def configure_optimizers(self):

    optimizer = AdamW(self.parameters(), lr=2e-5)

    scheduler = get_linear_schedule_with_warmup(
      optimizer,
      num_warmup_steps=self.n_warmup_steps,
      num_training_steps=self.n_training_steps
    )

    return dict(
        optimizer=optimizer,
        lr_scheduler=dict(scheduler=scheduler,interval='step'),
        
    )



max_token_length = 1024

#trained_model = IronicCommentModule.load_from_checkpoint('D:/noBert/LongNorBERT_Models/v1/epoch=9-step=4150.ckpt', n_classes=len(LABEL_COLUMNS))
trained_model = IronicCommentModule.load_from_checkpoint('D:/noBert/LongNorBERT_Models/v2/MODEL_NAME', n_classes=len(LABEL_COLUMNS))
#print(trained_model)

trained_model.eval()
trained_model.freeze()

tokenizer = BertTokenizerFast.from_pretrained('D:/noBert/LongNorBERT_v2')




train_data_path = 'D:\\noBert\\validation_dataset.csv'
train = pd.read_csv(train_data_path)


for i in range(0,train.shape[0]):
    frame = train.iloc[i]
    
    article_text = frame['article']
    comment_text = frame['comment']
    reply_text = frame['reply']
    
    if not str(reply_text) == 'nan':
        text = article_text.rstrip() +' [SEP] '+ comment_text.rstrip() +' [SEP] '+reply_text.rstrip()
    else:
        text = article_text.rstrip() +' [SEP] '+ comment_text.rstrip()

    train.at[i,'text'] = text



train = train[['text', 'label']].reset_index(drop=True)



#Remove too long
for i in range(0,train.shape[0]):
    frame = train.iloc[i]

    if len(tokenizer(frame['text'])[0]) > max_token_length:
        train.at[i,'text'] = 'nan'
        train.at[i,'label'] = 'nan'

train.dropna(inplace=True)


if False:
    #Investigate data
    for i in range(0,train.shape[0]):
        frame = train.iloc[i]
        print(frame['text'], frame['label'])
        print()
    input()
        
print(str(train['label'].value_counts()[1.0])+'/'+str(len(train['label'])))
print(str((train['label'].value_counts()[1.0]/len(train['label']))*100),'Percent are ironic')
print('\nBegin Evaluation\n')



if False:
    hits = []
    misses = []

    results = pd.DataFrame(columns=['Label','Prediction','Replies','Text'])
    #print(results)
    #input()

    #0.01
    for i in range(0,train.shape[0]):
        frame = train.iloc[i]
        
        encoding = tokenizer(
            frame['text'],
            max_length=max_token_len,
            return_token_type_ids=False,
            padding="max_length",
            truncation=True,
            return_attention_mask=True,
            return_tensors='pt',
            )
        
        _, test_prediction = trained_model(encoding["input_ids"], encoding["attention_mask"])
        pred = test_prediction.flatten().numpy()[0]

        temp = pd.DataFrame([[frame['label'], pred, frame['text'].split(' [SEP] ')[1:], frame['text']]], columns=['Label','Prediction','Replies','Text'])
        results = pd.concat([results, temp])
        #print(results)


        if pred > 0.01:

            if frame['label'] == 1.0:
                hits.append(frame['text'].split(' [SEP] ')[1:])
            else:
                misses.append(frame['text'].split(' [SEP] ')[1:])
                
        print(str(i+1)+'/290')
        
    print(len(hits))
    print(len(misses))

    results.to_csv('validation_prediction_results_long.csv', encoding='utf-8-sig', index=False)


















