#Gets model predictions and saves them to a CSV file to prevent having to get predicitions every time a metric is calculated.

import pandas as pd
import numpy as np
from tqdm.auto import tqdm
import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
from transformers import BertTokenizerFast, AutoTokenizer, BertModel, get_linear_schedule_with_warmup, BertForSequenceClassification
import pytorch_lightning as pl
from torch.optim import AdamW
from torchmetrics.functional import accuracy, f1, auroc
from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping
from pytorch_lightning.loggers import TensorBoardLogger
from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report, multilabel_confusion_matrix
from sklearn import metrics
import seaborn as sns
from pylab import rcParams
import os


os.environ["CUDA_VISIBLE_DEVICES"]="1"
os.environ["no_deprecation_warning"]="True"

LABEL_COLUMNS = ['label']
max_token_len = 512
max_token_length = 512


class ToxicCommentTagger(pl.LightningModule):

  def __init__(self, n_classes: int, n_training_steps=None, n_warmup_steps=None):
    super().__init__()
    self.bert = BertModel.from_pretrained('D:/noBert/norBERT2', return_dict=True)
    #self.bert = BertForSequenceClassification.from_pretrained('D:/noBert/norBERT2', return_dict=True)
    self.classifier = nn.Linear(self.bert.config.hidden_size, n_classes)
    self.n_training_steps = n_training_steps
    self.n_warmup_steps = n_warmup_steps
    self.criterion = nn.BCELoss()

  def forward(self, input_ids, attention_mask, label=None):
    output = self.bert(input_ids, attention_mask=attention_mask)
    output = self.classifier(output.pooler_output)
    output = torch.sigmoid(output)    
    loss = 0
    if label is not None:
        loss = self.criterion(output, label)
    return loss, output

  def training_step(self, batch, batch_idx):
    input_ids = batch["input_ids"]
    attention_mask = batch["attention_mask"]
    label = batch["label"]
    loss, outputs = self(input_ids, attention_mask, label)
    self.log("train_loss", loss, prog_bar=True, logger=True)
    return {"loss": loss, "predictions": outputs.detach(), "label": label}

  def validation_step(self, batch, batch_idx):
    input_ids = batch["input_ids"]
    attention_mask = batch["attention_mask"]
    label = batch["label"]
    loss, outputs = self(input_ids, attention_mask, label)
    self.log("val_loss", loss, prog_bar=True, logger=True)
    return loss

  def test_step(self, batch, batch_idx):
    input_ids = batch["input_ids"]
    attention_mask = batch["attention_mask"]
    label = batch["label"]
    loss, outputs = self(input_ids, attention_mask, label)
    self.log("test_loss", loss, prog_bar=True, logger=True)
    return loss

  def training_epoch_end(self, outputs):
    
    label = []
    predictions = []
    for output in outputs:
      for out_label in output["label"].detach().cpu():
        label.append(out_label)
      for out_predictions in output["predictions"].detach().cpu():
        predictions.append(out_predictions)

    label = torch.stack(label).int()
    predictions = torch.stack(predictions)

    for i, name in enumerate(LABEL_COLUMNS):
      class_roc_auc = auroc(predictions[:, i], label[:, i])
      self.logger.experiment.add_scalar(f"{name}_roc_auc/Train", class_roc_auc, self.current_epoch)


  def configure_optimizers(self):

    optimizer = AdamW(self.parameters(), lr=2e-5)

    scheduler = get_linear_schedule_with_warmup(
      optimizer,
      num_warmup_steps=self.n_warmup_steps,
      num_training_steps=self.n_training_steps
    )

    return dict(
      optimizer=optimizer,
      lr_scheduler=dict(
        scheduler=scheduler,
        interval='step'
      )
    )







trained_model = ToxicCommentTagger.load_from_checkpoint('D:/noBert/ShortNorBERT_Models/v3/MODEL_NAME.ckpt', n_classes=len(LABEL_COLUMNS))

trained_model.eval()
trained_model.freeze()

tokenizer = BertTokenizerFast.from_pretrained('D:/noBert/norBERT2')


train = pd.read_csv('D:\\noBert\\validation_dataset.csv')



duplicate = []

#Seperate rows
for i in range(0,train.shape[0]):
    frame = train.iloc[i]
    
    comment_text = frame['comment']
    reply_text = frame['reply']
    
    if not str(reply_text) == 'nan':
        text = reply_text#.rstrip()
    else:
        text = comment_text#.rstrip()

    if text in duplicate:
        continue
    duplicate.append(text)
    train.at[i,'text'] = text
    


train = train[['text', 'label']].reset_index(drop=True)


#Remove too long
for i in range(0,train.shape[0]):
    frame = train.iloc[i]
    #print(frame['text'])
    #input()

    if str(frame['text']) == 'nan':
      continue

    if len(tokenizer(frame['text'])[0]) > max_token_length:
        train.at[i,'text'] = 'nan'
        train.at[i,'label'] = 'nan'


train.dropna(inplace=True)



if False:
    #Investigate data
    for i in range(0,train.shape[0]):
        frame = train.iloc[i]
        print(frame['text'], frame['label'])
        print()
    input()





if True:
  results = pd.DataFrame(columns=['Label','Prediction','Text'])

  for i in range(0,train.shape[0]):
      frame = train.iloc[i]
      
      encoding = tokenizer(
          frame['text'],
          max_length=max_token_len,
          return_token_type_ids=False,
          padding="max_length",
          truncation=True,
          return_attention_mask=True,
          return_tensors='pt',
          )
      
      _, test_prediction = trained_model(encoding["input_ids"], encoding["attention_mask"])
      pred = test_prediction.flatten().numpy()[0]

      temp = pd.DataFrame([[frame['label'], pred, frame['text']]], columns=['Label','Prediction','Text'])
      results = pd.concat([results, temp])

      print(str(i+1)+'/290')

  results.to_csv('validation_prediction_results_short.csv', encoding='utf-8-sig', index=False)









